/******************************************************************************

Copyright (c) 1997, Andrew J. Weber, Jason P. Weber

Copyright information should have been included with this distribution
and may be found at ../meta/COPYRIGHT or obtained via email from weber@acm.org

******************************************************************************/

#include <bld/checkop.h>
#include <bld/target.h>
#include <bld/interp.h>

/***********************************************************************
class BLD_BuildOperation
***********************************************************************/
const GPL_String BLD_BuildOperation::current = "CURRENT";
const GPL_String BLD_BuildOperation::built = "BUILT";
const GPL_String BLD_BuildOperation::failed = "FAILED";

BLD_Value BLD_BuildOperation::Perform(void) const
{
	BLD_Target::SetExecuted();

	GPL_String output = current;
	GPL_String result;
	BLD_Value targets;
	register const unsigned long n = ops.GetCount();
	for (register unsigned long i = 0; n > i; ++i)
	{
		targets = ops[i]->Perform();
		for (register unsigned long j = 0; targets.GetCount() > j; ++j)
		{
			Build(targets[j], result);
//WBA_APP->Printf("Build(%s) => %s", targets[j].GetData(), result.GetData());
			if (failed == result && (BLD_ARG_OFF("BLD_PERSIST") &&
								BLD_ARG_OFF("BLD_IGNORE_TARGET_FAILURES")))
			{
				return result;
			}
			Update(result, output);
		}
	}
	return output;
}

void BLD_BuildOperation::Update(const GPL_String & result, GPL_String & status)
{
	if (failed != status && current != result)
	{
		status = result;
	}
}

void BLD_BuildOperation::Build(const GPL_String & name, GPL_String & result)
{
	result = current;

	// If this is a null target...
	if (0 == name.GetLength())
	{
		// Ignore it.
		return;
	}

	// Check the target.
//WBA_APP->Printf("Building \"%s\"", name.GetData());
	switch (BLD_CONTEXT.Build(name))
	{
	case BLD_BUILD_UNNECESSARY:
		// NOP
		break;
	case BLD_BUILD_NEEDED:
		BLD_Target::SetExecuted();
		result = built;
		break;
	default:
		BLD_Target::SetFailed();
		result = failed;
		if (BLD_ARG_OFF("BLD_PERSIST") &&
			BLD_ARG_OFF("BLD_IGNORE_TARGET_FAILURES"))
		{
			unsigned long k = BLD_Target::targets.GetCount();
			while (1 < k)
			{
				BLD_Target::targets[--k]->Terminate();
			}
		}
		break;
	}
}

GPL_String BLD_BuildOperation::ToString(void) const
{
	GPL_String image = BLD_ListOperation::ToString();
	image.Prepend("build ");
	return image;
}

/***********************************************************************
class BLD_TargetCheckOperation
***********************************************************************/
BLD_Value BLD_TargetCheckOperation::Perform(void) const
{
	GPL_String output = BLD_BuildOperation::current;
	GPL_String result = BLD_BuildOperation::current;
	BLD_Value targets;
	register const unsigned long n = ops.GetCount();
	for (register unsigned long i = 0; n > i; ++i)
	{
		targets = ops[i]->Perform();
		for (register unsigned long j = 0; targets.GetCount() > j; ++j)
		{
			Check(targets[j], descend, result);
			if (BLD_BuildOperation::failed == result)
			{
				output = result;
			}
			if (BLD_ARG_OFF("BLD_PERSIST") &&
				BLD_ARG_OFF("BLD_IGNORE_TARGET_FAILURES"))
			{
				break;
			}
			if (BLD_BuildOperation::current == output)
			{
				output = result;
			}
		}
	}
	return output;
}

void BLD_TargetCheckOperation::Check(const GPL_String & name,
									const GPL_Boolean descend,
									GPL_String & result)
{
	result = BLD_BuildOperation::current;

	// If this is a null target...
	if (0 == name.GetLength())
	{
		// Ignore it.
		return;
	}

	// Check the target.
	const BLD_BuildStatus status = BLD_CONTEXT.Check(name, descend);
//WBA_APP->Printf("Check(\"%s\", %s) => %d", name.GetData(), (descend ? "TRUE" : "FALSE"), status);
	switch (status)
	{
	case BLD_BUILD_UNNECESSARY:
		if (!BLD_STATE(BLD_Target::targets.Top()->GetName()).HasTimeStamp() ||
			BLD_STATE(BLD_Target::targets.Top()->GetName()).GetLastBuilt() < BLD_STATE(name).GetLastBuilt())
		{
//WBA_APP->Printf("\n%s%s\n%s%s",
//BLD_Target::targets.Top()->GetName().GetData(),
//BLD_STATE(BLD_Target::targets.Top()->GetName()).ToString().GetData(),
//name.GetData(),
//BLD_STATE(name).ToString().GetData());
			result = BLD_BuildOperation::built;
		}
		break;

	case BLD_BUILD_NEEDED:
		BLD_Target::SetExecuted();
		result = BLD_BuildOperation::built;
		break;

	default:
		BLD_Target::SetFailed();
		result = BLD_BuildOperation::failed;
		if (BLD_ARG_OFF("BLD_PERSIST") &&
			BLD_ARG_OFF("BLD_IGNORE_TARGET_FAILURES"))
		{
			unsigned long k = BLD_Target::targets.GetCount();
			while (1 < k)
			{
				BLD_Target::targets[--k]->Terminate();
			}
		}
		break;
	}
}

GPL_String BLD_TargetCheckOperation::ToString(void) const
{
	GPL_String image = BLD_ListOperation::ToString();
	image.Prepend(descend ? "tcheck(" : "scheck(");
	image.Append(")");
	return image;
}

/***********************************************************************
class BLD_FileCheckOperation
***********************************************************************/
BLD_Value BLD_FileCheckOperation::Perform(void) const
{
	GPL_String output = BLD_BuildOperation::current;
	BLD_Value targets;
	register const unsigned long n = ops.GetCount();
	for (register unsigned long i = 0; n > i; ++i)
	{
		targets = ops[i]->Perform();
		for (register unsigned long j = 0; targets.GetCount() > j; ++j)
		{
			if (IsNew(targets[j]))
			{
				BLD_BuildOperation::Update(BLD_BuildOperation::built, output);
				break;
			}
		}
	}
	return output;
}

GPL_Boolean BLD_FileCheckOperation::IsNew(const GPL_String & name)
{
	// If this is a null target...
	if (0 == name.GetLength())
	{
		// Ignore it.
		return FALSE;
	}

	// If the reference does not exist...
	BLD_Target *t = (BLD_Target *) GPL_BAD_ADDRESS;
	if (!BLD_Target::targets.Peek(t))
	{
		return FALSE;
	}
	if ((BLD_Target *) GPL_BAD_ADDRESS == t)
	{
		return FALSE;
	}
	const GPL_String reference = t->GetName();
	GPL_Time referenceTime;
	if (!BLD_APP->GetModificationTime(reference, referenceTime))
	{
		if (BLD_ARG_ON("BLD_DEBUG"))
		{
			WBA_APP->Printf("Unable to get modification time for reference %s",
							reference.GetData());
		}
		return FALSE;
	}

	// If the target does not exist...
	GPL_Time targetTime;
	if (!BLD_APP->GetModificationTime(name, targetTime))
	{
		if (BLD_ARG_ON("BLD_DEBUG"))
		{
			WBA_APP->Printf("Unable to get modification time for target %s",
							name.GetData());
		}
		return FALSE;
	}

	// If the target file has been modified more
	// recently than the reference file...
	if (targetTime > referenceTime)
	{
		if (BLD_ARG_ON("BLD_DEBUG"))
		{
			WBA_APP->Printf("%s is newer than %s", name.GetData(),
							reference.GetData());
		}
		return TRUE;
	}

	return FALSE;
}

GPL_String BLD_FileCheckOperation::ToString(void) const
{
	GPL_String image = BLD_ListOperation::ToString();
	image.Prepend("fcheck(");
	image.Append(")");
	return image;
}

/***********************************************************************
class BLD_CheckOperation
***********************************************************************/
BLD_Value BLD_CheckOperation::Perform(void) const
{
	GPL_String output = BLD_BuildOperation::current;
	GPL_String result = BLD_BuildOperation::current;
	BLD_Value targets;
	for (register unsigned long i = 0; ops.GetCount() > i; ++i)
	{
		targets = ops[i]->Perform();
		for (register unsigned long j = 0; targets.GetCount() > j; ++j)
		{
//WBA_APP->Printf("targets[%d,%d] == \"%s\"", i, j, targets[j].GetData());
			if (BLD_FileCheckOperation::IsNew(targets[j]))
			{
				BLD_BuildOperation::Update(BLD_BuildOperation::built, output);
				if (BLD_ARG_ON("BLD_DEBUG"))
				{
					WBA_APP->Printf("File %s forcing rebuild",
										targets[j].GetData());
				}
			}
			BLD_TargetCheckOperation::Check(targets[j], TRUE, result);
			if (BLD_BuildOperation::built == result &&
				BLD_ARG_ON("BLD_DEBUG"))
			{
				WBA_APP->Printf("Target %s forcing rebuild",
									targets[j].GetData());
			}
			BLD_BuildOperation::Update(result, output);
			if (BLD_BuildOperation::failed == output &&
				BLD_ARG_OFF("BLD_PERSIST") &&
				BLD_ARG_OFF("BLD_IGNORE_TARGET_FAILURES"))
			{
				break;
			}
		}
	}
	return output;
}

GPL_String BLD_CheckOperation::ToString(void) const
{
	GPL_String image = BLD_ListOperation::ToString();
	image.Prepend("check(");
	image.Append(")");
	return image;
}
