#ifndef BLD_OP_H
#define BLD_OP_H

/******************************************************************************

Copyright (c) 1997, Andrew J. Weber, Jason P. Weber

Copyright information should have been included with this distribution
and may be found at ../meta/COPYRIGHT or obtained via email from weber@acm.org

******************************************************************************/

#include <gpl/gpl.h>
#include <bld/value.h>

/***********************************************************************
class BLD_Operation

Abstract base class for all operations.
***********************************************************************/
class BLD_Operation
{
	public:
		virtual ~BLD_Operation(void) {}

		virtual BLD_Value Perform(void) const { return BLD_Value(); }
		virtual GPL_String ToString(void) const { return ""; }
};

typedef	GPL_Array< GPL_Pointer<BLD_Operation> > OpArray;

/***********************************************************************
class BLD_LiteralOperation

Literal string with no interpretation.
***********************************************************************/
class BLD_LiteralOperation : public BLD_Operation
{
	public:
		BLD_LiteralOperation(void) {}
		BLD_LiteralOperation(const char c) : string(1, c) {}
		BLD_LiteralOperation(const char *s) : string(s) {}
		virtual ~BLD_LiteralOperation(void) {}

		virtual BLD_Value Perform(void) const;
		virtual GPL_String ToString(void) const;

	protected:
		GPL_String string;
};

/***********************************************************************
class BLD_StringOperation

String with interpretation.
***********************************************************************/
class BLD_StringOperation : public BLD_LiteralOperation
{
	public:
		BLD_StringOperation(const char *s) : BLD_LiteralOperation(s) {}
		virtual ~BLD_StringOperation(void) {}

		virtual BLD_Value Perform(void) const;
		virtual GPL_String ToString(void) const;
};

/***********************************************************************
class BLD_CompositeOperation

Composite of operations.
***********************************************************************/
class BLD_CompositeOperation : public BLD_Operation
{
	public:
		BLD_CompositeOperation(void) {}
		BLD_CompositeOperation(const GPL_Pointer<BLD_Operation> & o);
		BLD_CompositeOperation(const OpArray & o)
			: ops(o) {}
		virtual ~BLD_CompositeOperation(void) {}

		virtual void Add(const GPL_Pointer<BLD_Operation> & o);

		virtual BLD_Value Perform(void) const;
		virtual GPL_String ToString(void) const;

	protected:
		OpArray ops;
};

/***********************************************************************
class BLD_ListOperation

List of operations.
***********************************************************************/
class BLD_ListOperation : public BLD_CompositeOperation
{
	public:
		BLD_ListOperation(void) {}
		BLD_ListOperation(const GPL_Pointer<BLD_Operation> & o)
			: BLD_CompositeOperation(o) {}
		BLD_ListOperation(const OpArray & o)
			: BLD_CompositeOperation(o) {}
		virtual ~BLD_ListOperation(void) {}

		virtual BLD_Value Perform(void) const;
		virtual GPL_String ToString(void) const;
};

/***********************************************************************
class BLD_BlockOperation

Block of operations.
***********************************************************************/
class BLD_BlockOperation : public BLD_ListOperation
{
	public:
		BLD_BlockOperation(const OpArray & o)
			: BLD_ListOperation(o) {}
		virtual ~BLD_BlockOperation(void) {}

		virtual BLD_Value Perform(void) const;
		virtual GPL_String ToString(void) const;

	protected:
		BLD_BlockOperation(void) {}
};

#endif /* BLD_OP_H */
